/*************************************************************************
 *                                                                       *
 * GLE - Graphics Layout Engine <http://www.gle-graphics.org/>           *
 *                                                                       *
 * BSD License                                                           *
 *                                                                       *
 * Copyright (c).                                                        *
 * All rights reserved.                                                  *
 *                                                                       *
 * Redistribution and use in source and binary forms, with or without    *
 * modification, are permitted provided that the following conditions    *
 * are met:                                                              *
 *                                                                       *
 *  * Redistributions of source code must retain the above copyright     *
 *  notice, this list of conditions and the following disclaimer.        *
 *                                                                       *
 *  * Redistributions in binary form must reproduce the above            *
 *  copyright notice, this list of conditions and the following          *
 *  disclaimer in the documentation and/or other materials provided      *
 *  with the distribution.                                               *
 *                                                                       *
 *  * Neither the name of GLE nor the names of its contributors may      *
 *  be used to endorse or promote products derived from this software    *
 *  without specific prior written permission.                           *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT      *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT   *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  *
 *                                                                       *
 *************************************************************************/

#include "all.h"
#include "tokens/stokenizer.h"
#include "SourceLine.h"
#include "tokens/Tokenizer.h"
#include "mem_limits.h"
#include "token.h"
#include "core.h"
#include "var.h"
#include "glearray.h"
#include "polish.h"
#include "pass.h"
#include "file_io.h"
#include "gprint.h"
#include "mygraph.h"
#include "run.h"
#include "cmdline.h"
#include "config.h"
#include "drawit.h"
#include "cutils.h"
#include "gleinterface.h"

#define dbg if (gle_debug>0)
extern int gle_debug;
extern int this_line;
int trace_on;
int *(*gpcode)[];   /* gpcode is a pointer to an array of poiter to int */
int (*gplen)[];   /* gpcode is a pointer to an array of int */
int ngpcode=0;
int ngerror;
extern int last_line;
void expand_pcode(int i, int *j);
char *dr_nextline(int *srclin);
void dr_init(void);
int getch(void);
int abort_flag;
string g_OutputFile;

void g_close();
int gle_is_open();

extern CmdLineObj g_CmdLine;

const string& get_output_file() {
	return g_OutputFile;
}

void do_set_vars() {
	var_def("PI", GLE_PI);
	var_def("XGMIN", 0.0);
	var_def("YGMIN", 0.0);
	var_def("XGMAX", 0.0);
	var_def("YGMAX", 0.0);
	var_def("X2GMIN", 0.0);
	var_def("Y2GMIN", 0.0);
	var_def("X2GMAX", 0.0);
	var_def("Y2GMAX", 0.0);
	var_def("ZGMIN", 0.0);
	var_def("ZGMAX", 0.0);
}

/*---------------------------------------------------------------------------*/

void output_error(ParserError& err) {
	g_set_error_column(-1);
	if (err.hasFlag(TOK_PARSER_ERROR_ATEND)) {
		// Otherwise, the message would be "unexpected end of file"
		err.setMessage("unexpected end of line");
	}
	if (err.hasFlag(TOK_PARSER_ERROR_PSTRING)) {
		if (err.file() == "") {
			gprint(string(">> Error: ")+err.msg()+"\n");
		} else {
			string err_str;
			err.toString(err_str);
			gprint(string(">> Error: ")+err_str+"\n");
		}
		if (err.getColumn() != -1) {
			stringstream pos_strm;
			pos_strm << ">> In: '";
			int delta = showLineAbbrev(err.getParserString(), err.getColumn(), pos_strm);
			pos_strm << "'" << endl;
			pos_strm << ">>";
			for (int i = 0; i < err.getColumn()+5-delta; i++) {
				pos_strm << " ";
			}
			pos_strm << "^" << endl;
			gprint(pos_strm.str());
		}
	} else {
		if (err.file() == "") {
			// Error in the GLE file
			g_set_error_column(err.getColumn());
			gprint(string(">> Error: ")+err.msg()+"\n");
		} else {
			// Error while reading a data file
			string err_str;
			err.toString(err_str);
			gprint(string(">> Error: ")+err_str+"\n");
		}
	}
}

void output_error_cerr(ParserError& err) {
	// Used by GLE "as a calculator", i.e., if no GLE file is active
	if (err.hasFlag(TOK_PARSER_ERROR_ATEND)) {
		// Otherwise, the message would be "unexpected end of file"
		err.setMessage("unexpected end of line");
	}
	if (err.hasFlag(TOK_PARSER_ERROR_PSTRING)) {
		cerr << ">> Error: " << err.msg() << endl;
		if (err.getColumn() != -1) {
			cerr << ">> In: '" << err.getParserString() << "'" << endl;
			stringstream pos_strm;
			pos_strm << ">>";
			for (int i = 0; i < err.getColumn()+5; i++) {
				pos_strm << " ";
			}
			pos_strm << "^" << endl;
			cerr << pos_strm.str();
		}
	} else {
		cerr << ">> Error: " << err.msg() << endl;
	}
}

void validate_file_name(const string& fname, bool isread) throw(ParserError) {
	GLEGlobalConfig* conf = GLEGetInterfacePointer()->getConfig();
	if (conf->getCmdLine()->hasOption(GLE_OPT_SAFEMODE)) {
		bool allow = false;
		string fullpath, dirname;
		GLEGetCrDir(&dirname);
		GLEGetFullPath(dirname, fname, fullpath);
		// cout << "full path: " << fullpath << endl;
		GetDirName(fullpath, dirname);
		StripDirSepButNotRoot(dirname);
		int nbread = conf->getNumberAllowReadDirs();
		if (nbread > 0 && isread) {
			for (int i = 0; i < nbread; i++) {
				if (conf->getAllowReadDir(i) == dirname) {
					allow = true;
				}
			}
			if (allow) {
				return;
			} else {
				g_throw_parser_error("safe mode - reading not allowed in directory '", dirname.c_str(), "'");
			}
		}
		int nbwrite = conf->getNumberAllowWriteDirs();
		if (nbwrite > 0 && !isread) {
			for (int i = 0; i < nbwrite; i++) {
				if (conf->getAllowWriteDir(i) == dirname) {
					allow = true;
				}
			}
			if (allow) {
				return;
			} else {
				g_throw_parser_error("safe mode - writing not allowed in directory '", dirname.c_str(), "'");
			}
		}
		g_throw_parser_error("safe mode - can not access '", fname.c_str(), "': file system access has been disabled");
	}
}

void DrawIt(const string& output_file, GLEGlobalSource &glecode, CmdLineObj* cmdline, bool silent) throw(ParserError) {
	abort_flag = false;
	ngerror = 0;
	last_line = 0;
	if (!silent) {
		string version;
		g_get_version(&version);
		printf("GLE %s [%s]-C", version.c_str(), glecode.getFileName().c_str());
		fflush(stdout);
	}
	g_OutputFile = output_file;
	g_clear();
	var_clear();
	mark_clear();
	sub_clear();
	name_clear();
	clear_run();
	f_init();
	if (cmdline != NULL) {
		int devtype = g_get_device();
		if (devtype == GLE_DEVICE_PS) {
			g_set_fullpage(true);
		} else {
			g_set_fullpage(cmdline->hasOption(GLE_OPT_FULL_PAGE));
		}
	}
	g_set_pagesize(gle_config_papersize());
	g_set_margins(gle_config_margins());
	do_set_vars();

	// Auto-include "compatibility.gle" if compatibilty mode is not most recent
	string compat = "compatibility.gle";
	if ((g_get_compatibility() < GLE_COMPAT_MOST_RECENT) && !glecode.includes(compat)) {
		GLESourceFile* includefile = new GLESourceFile();
		text_load_include(compat, includefile);
		glecode.insertIncludeNoOverwrite(0, includefile);
	}

	GLEPcodeList pc_list;
	GLEPcode pcode(&pc_list);

	int maxpcode=0;
	expand_pcode(glecode.getNbLines(),&maxpcode);
	ngpcode=0;

	GLEPolish polish;
	polish.initTokenizer();

	// Create tokenizer
	GLEParser parser(&polish);
	parser.initTokenizer();

	set_global_parser(&parser);

	for (int crline = 0; crline < glecode.getNbLines(); crline++) {
		// call passt to convert the tokens into PCode
		// pcode is simply numbers that instruct the drawing
		// engine which commands to draw

		// cout << "Line = " << *crline;
		GLESourceLine* code = glecode.getLine(crline);
		parser.setString(code->getCodeCStr());

		try {
			parser.passt(*code, pcode);
		} catch (ParserError err) {
			output_error(err);
		}

		bool add_pcode = true;
		if (parser.hasSpecial(GLE_PARSER_INCLUDE) && !glecode.includes(parser.getInclude())) {
			// this is a file to include load the file and insert it into glecode here
			GLESourceFile* includefile = new GLESourceFile();
			text_load_include(parser.getInclude(), includefile);
			glecode.insertInclude(crline, includefile);
			crline--;
			if (g_verbosity() > 5) {
				cout << "{" << parser.getInclude() << "}";
			}
			// don't need this pcode
			add_pcode = false;
		}
		if (add_pcode){
			// copy the pcode into the global buffer
			if (ngpcode > maxpcode) expand_pcode(ngpcode*2,&maxpcode);
			(*gpcode)[++ngpcode] = pcode.size() == 0 ? NULL : (int*) myallocz(pcode.size()*sizeof(char *));
			(*gplen)[ngpcode] = pcode.size();
			memcpy((*gpcode)[ngpcode],&pcode[0],pcode.size()*sizeof(int));
		}
		pcode.clear();
	}
	try {
		parser.checkmode();
	} catch (ParserError err) {
		output_error(err);
	}
	if (ngerror > 0){
		reset_new_error(true);
		g_message("");
		g_throw_parser_error("errors, aborting");
	}
	g_compatibility_settings();
	try {
		//
		// -- now run the pcode in the driver
		//
		if (!silent) printf("-R-");
		if (ngpcode != glecode.getNbLines()) {
			cout << "error pcode and text size mismatch"<<endl;
			cout << "pcode size = " << ngpcode <<" text size = "<<glecode.getNbLines()<<endl;
		}
		token_space();
		int endp = 0;
		// should get rid of glecode here!
		GLEInterface* iface = GLEGetInterfacePointer();
		bool mkdrobjs = iface->isMakeDrawObjects();
		for (int i=1;i<=ngpcode;i++){
			this_line = i;
			GLESourceLine* code = glecode.getLine(i-1);
			do_pcode(*code,&i,(*gpcode)[i],(*gplen)[i],&endp,mkdrobjs);
		}
	} catch (ParserError err) {
		output_error(err);
	}
	if (!gle_is_open()) {
		// If .gle file is empty, create empty postscript file
		if (!g_has_size()) g_set_size(10, 10, false);
		g_open(get_output_file(), glecode.getFileName());
	}
	set_global_parser(NULL);
	bool has_console = g_reset_message();
	g_close();
	g_set_console_output(has_console);
}

void text_load_include(const string& fname, GLESourceFile* file) throw(ParserError) {
	string fullname = GetActualFilename(fname);
	if (fullname == "") {
		g_throw_parser_error("file not found: '", fname.c_str(), "'");
	} else {
		validate_file_name(fullname, true);
		text_load(fullname, fname, file);
		file->trim(0);
	}
}

void text_load(const string& fname, GLESourceFile* file) throw(ParserError) {
	text_load(fname, fname, file);
}

bool text_load(const string& fullname, GLEGlobalSource& file) throw(ParserError) {
	GLESourceFile* main = new GLESourceFile();
	file.setMainFile(main);
	bool res = text_load(fullname, fullname, main, true);
	main->trim(1);
	file.initFromMain();
	return res;
}

bool text_load(const string& fullname, const string& fname, GLESourceFile* file, bool failOnErr) throw(ParserError) {
	// -- load the gle file called by main program and the include command
	//    fname .. the filename
	//    SL    .. STL list of source lines
	// returns true upon success false if it cant find the file
	//
	if (IsDirectory(fullname)) {
		if (failOnErr) return false;
		g_throw_parser_error("file is a directory: '", fname.c_str(), "'");
	}
	std::ifstream input;
	input.open(fullname.c_str());
	if (!input.is_open()){
		if (failOnErr) return false;
		g_throw_parser_error("file not found: '", fname.c_str(), "'");
	}
	//
	// -- load the input and rememeber to look for continuation character
	//
	const char CONT_CHAR = '&';  // should be a global define somewhere??
	bool cont = false;
	string inbuff;
	file->setFileName(fname);
	while(!input.eof()){
		string linbuff;
		getline(input, linbuff);
		// trim is required to make position indicator in error messages appear correctly
		str_trim_right(linbuff);
		if (cont) {
			// -- splice this string into inbuff at location of '&'
			str_trim_left(linbuff);
			inbuff.replace(inbuff.rfind(CONT_CHAR),linbuff.length(),linbuff);
			cont = false;
		} else {
			inbuff = linbuff;
		}
		// -- search for continuation character '&' if found
		if (inbuff.length() > 0 && inbuff.at(inbuff.length()-1) == CONT_CHAR) {
			// -- continuation
			cont = true;
		}
		if (!cont || input.eof()){
			string prefix;
			GLESourceLine* sline = file->addLine();
			str_trim_left(inbuff, prefix);
			sline->setPrefix(prefix);
			sline->setCode(inbuff);
		}
	}
	input.close();
	return true;
}

void expand_pcode(int ngtxt,int *maxpcode) {
        int *a,*b;
        a = (int*) myallocz((ngtxt+10)*sizeof(char *));
        b = (int*) myallocz((ngtxt+10)*sizeof(char *));
        if (gpcode!=0) {
                memcpy(a,gpcode,(*maxpcode + 4)*sizeof(char *));
                memcpy(b,gplen,(*maxpcode + 4)*sizeof(char *));
                myfrees(gpcode,"gpcode2");
                myfrees(gplen,"gplen");
        }
        gpcode = (int *(*)[]) a;
        gplen = (int (*)[]) b;
        *maxpcode = ngtxt+1;
}

