/*************************************************************************
 *                                                                       *
 * GLE - Graphics Layout Engine <http://www.gle-graphics.org/>           *
 *                                                                       *
 * BSD License                                                           *
 *                                                                       *
 * Copyright (c).                                                        *
 * All rights reserved.                                                  *
 *                                                                       *
 * Redistribution and use in source and binary forms, with or without    *
 * modification, are permitted provided that the following conditions    *
 * are met:                                                              *
 *                                                                       *
 *  * Redistributions of source code must retain the above copyright     *
 *  notice, this list of conditions and the following disclaimer.        *
 *                                                                       *
 *  * Redistributions in binary form must reproduce the above            *
 *  copyright notice, this list of conditions and the following          *
 *  disclaimer in the documentation and/or other materials provided      *
 *  with the distribution.                                               *
 *                                                                       *
 *  * Neither the name of GLE nor the names of its contributors may      *
 *  be used to endorse or promote products derived from this software    *
 *  without specific prior written permission.                           *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT      *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT   *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  *
 *                                                                       *
 *************************************************************************/

#include <math.h>
#include <time.h>
#include <string>
#include <sstream>
#include <iostream>
#include <vector>
#include <list>
#include <map>
#include <fstream>
#include <algorithm>

#include "basicconf.h"

#include <sys/stat.h>

#ifdef __WIN32__
	#include <sys/types.h>
	#include <sys/stat.h>
	#include <stdlib.h>
	#include <process.h>
	#include <windows.h>
	#include <winsock.h>
#endif
#ifdef __CYGWIN__
	#include <unistd.h>
	#include <windows.h>
#endif
#ifdef __MACOS__
	#include <sys/param.h>
	#include <mach-o/dyld.h>
#endif
#ifdef __UNIX__
	#include <limits.h>
	#include <unistd.h>
	#include <sys/types.h>
	#include <sys/wait.h>
	#ifdef HAVE_SOCKETS
		#include <sys/socket.h>
		#include <netinet/in.h>
		#include <arpa/inet.h>
		typedef int SOCKET;
		#define SOCKET_ERROR -1
	#endif
#endif
#if defined(__OS2__) && defined(__EMX__)
	#include <unistd.h>
	#include <sys/types.h>
	#ifdef HAVE_SOCKETS
		#include <sys/socket.h>
		#include <netinet/in.h>
		#include <arpa/inet.h>
		typedef int SOCKET;
		#define SOCKET_ERROR -1
	#endif
#endif

using namespace std;

#include "tokens/stokenizer.h"
#include "errno.h"
#include "stdio.h"
#include "stdlib.h"
#include "file_io.h"
#include "cutils.h"

#ifdef __UNIX__
	/* Cygwin too */
	string PATH_SEP = ":";
	string DIR_SEP = "/";
#endif

#ifdef __WIN32__
	string PATH_SEP = ";";
	string DIR_SEP = "\\";
#endif

#if defined(__OS2__) && defined(__EMX__)
	string PATH_SEP = ";";
	string DIR_SEP = "/";
#endif

string GLE_TOP_DIR;
string GLE_BIN_DIR;

//
// -- directory finding routines
//

void FillIncludePaths(vector<string> &IP) {
	// fills containor with paths to search for
	// file.  This is called if it can't find file
	// in current directory
	//
	// Paths searched are:
	// %GLE_TOP%
	// %GLE_TOP%/gleinc
	// %GLE_USRLIB%
	// note that GLE_USERLIB can contain multiple
	// directories separated by ; (PC) or : (unix)
	// as defined in PATH_SEP
	//
	//printf("string paths for GLE_PATH\n");
	string paths = GLE_TOP_DIR + DIR_SEP + "gleinc";
	IP.push_back(paths);
	//printf("paths for GLE_USRLIB\n");
	const char* usr_lib = getenv("GLE_USRLIB");
	if (usr_lib != NULL) {
		paths = getenv("GLE_USRLIB");
		GLEPathToVector(paths, &IP);
	}
}

void GLEPathToVector(const string& path, vector<string>* vec) {
	char_separator sep(PATH_SEP.c_str());
	tokenizer<char_separator> tokens(path, sep);
	while (tokens.has_more()) {
		string dir = tokens.next_token();
		CorrectDirSepStrip(dir);
		vec->push_back(dir);
	}
}

void StripDirSep(string& fname) {
	if (str_i_ends_with(fname, DIR_SEP.c_str())) {
		int nb = DIR_SEP.length();
		fname.erase(fname.length()-nb, nb);
	}
}

void StripDirSepButNotRoot(string& fname) {
	if (str_i_ends_with(fname, DIR_SEP.c_str())) {
		if (fname == "/") {
			// don't strip unix root dir
			return;
		}
		int nb = DIR_SEP.length();
		fname.erase(fname.length()-nb, nb);
	}
}

void CorrectDirSep(string& fname) {
	int len = fname.length();
	char dirsep = DIR_SEP[0];
	for (int i = 0; i < len; i++) {
		if (fname[i] == '/' || fname[i] == '\\') {
			fname[i] = dirsep;
		}
	}
}

void CorrectDirSepStrip(string& fname) {
	CorrectDirSep(fname);
	StripDirSepButNotRoot(fname);
}

void AddDirSep(string& fname) {
	if (!str_i_ends_with(fname, DIR_SEP.c_str())) {
		fname += DIR_SEP;
	}
}

bool IsAbsPath(const string& path) {
	if (path.length() >= 1) {
		if (path[0] == '/') {
			// Unix style absolute path
			return true;
		}
		if (path.length() >= 3) {
			if (path[1] == ':' && (path[2] == '/' || path[2] == '\\')) {
				// Windows style absolute path
				return true;
			}
		}
	}
	return false;
}

void GLEGetFullPath(const string& dirname, const string& fname, string& fullpath) {
	if (IsAbsPath(fname)) {
		fullpath = fname;
	} else {
		fullpath = dirname;
		AddDirSep(fullpath);
		fullpath += fname;
	}
	// search for "/../"
	string find = DIR_SEP + ".." + DIR_SEP;
	string::size_type i = fullpath.find(find);
	while (i != string::npos) {
		if (i == 0) return;
		// go to position before dir separator
		string::size_type j = i - 1;
		while (fullpath[j] != '/' && fullpath[j] != '\\') {
			if (j == 0) return;
			j--;
		}
		// erase from previous directory separator to end of "/../"
		fullpath.erase(j, i-j+find.length()-1);
		i = fullpath.find(find);
	}
}

// Get extension of filename and convert it to lower case
void GetExtension(const string& fname, string& ext) {
	string::size_type i = fname.rfind('.');
	if (i != string::npos) {
		ext = fname.substr(i+1);
		string::size_type len = ext.length();
		for (i = 0; i < len; i++) {
			int ch = ext.at(i);
			if (ch >= 'A' && ch <= 'Z') {
				ch += 'a' - 'A';
				ext.at(i) = ch;
			}
		}
	} else {
		ext = "";
	}
}

// Get file name without extension
void GetMainName(const string& fname, string& name) {
	string::size_type pos = fname.length();
	while (pos >= 1 && fname[pos-1] != '/' && fname[pos-1] != '\\' && fname[pos-1] != '.' ) {
		pos--;
	}
	if (pos >= 1 && fname[pos-1] == '.') {
		name = fname.substr(0, pos-1);
	} else {
		name = fname;
	}
}

// Add extension to filename
void AddExtension(string& fname, const string& ext) {
	string::size_type pos = fname.length();
	while (pos >= 1 && fname[pos-1] != '/' && fname[pos-1] != '\\' && fname[pos-1] != '.' ) {
		pos--;
	}
	if (pos >= 1 && fname[pos-1] == '.') {
		fname.erase(pos);
		fname += ext;
	} else {
		fname += ".";
		fname += ext;
	}
}

// Get path of file name (assume last part after DIR_SEP is file name)
void GetDirName(const string& fname, string& dir) {
	string::size_type i = fname.rfind(DIR_SEP);
	if (i != string::npos) {
		dir = fname.substr(0, i+1);
		AddDirSep(dir);
	} else {
		dir = "";
	}
}

// Split file name in path (dir) and file name (name)
void SplitFileName(const string& fname, string& dir, string& name) {
	string::size_type pos = fname.length();
	while (pos >= 1 && fname[pos-1] != '/' && fname[pos-1] != '\\') {
		pos--;
	}
	if (pos >= 1 && (fname[pos-1] == '/' || fname[pos-1] == '\\')) {
		dir = fname.substr(0, pos);
		name = fname.substr(pos);
		AddDirSep(dir);
	} else {
		name = fname;
		dir = "";
	}
}

// Split file name in path (dir) and file name (name) and return only file name
void SplitFileNameNoDir(const string& fname, string& name) {
	string::size_type i = fname.rfind(DIR_SEP);
	if (i != string::npos) {
		name = fname.substr(i+1);
	} else {
		name = fname;
	}
}

// Remove given absolute path from filename
void RemoveDirectoryIfEqual(string* filename, const string& directory) {
	if (IsAbsPath(directory)) {
		// Ignore trailing path separator on directory
		int pos = directory.length()-1;
		while (pos > 0 && (directory[pos] == '/' || directory[pos] == '\\')) pos--;
		// Check if directory is a prefix of the given filename
		if (strncmp(directory.c_str(), filename->c_str(), pos+1) == 0) {
			// Skip the path separator character
			pos++;
			// And erase the absolute path prefix
			if (pos < (int)filename->length() && ((*filename)[pos] == '/' || (*filename)[pos] == '\\')) {
				filename->erase(0, pos+1);
			}
		}
	}
}

void StripPathComponents(string* fname, int nb) {
	while (nb > 0) {
		string::size_type i = fname->rfind(DIR_SEP);
		if (i != string::npos) {
			*fname = fname->substr(0, i);
		} else {
			break;
		}
		nb--;
	}
}

bool GLEMoveFile(const string& from, const string& to) {
#ifdef __WIN32__
#else
	if (rename(from.c_str(), to.c_str()) == -1) return false;
#endif
	return true;
}

bool GLECopyFile(const string& from, const string& to) {
	ofstream out(to.c_str());
	ifstream strm(from.c_str());
	while (!strm.eof()) {
		char ch;
		strm.read(&ch, 1);
		out << ch;
	}
	out.close();
	strm.close();
	return true;
}

// Return current directory
bool GLEGetCrDir(string* name) {
#ifdef __WIN32__
	TCHAR buffer[1024];
	if (GetCurrentDirectory(1024, buffer) != 0) {
		*name = buffer;
		return true;
	} else {
		return false;
	}
#else
	char buffer[1024];
	if (getcwd(buffer, 1024) == NULL) {
		return false;
	} else {
		*name = buffer;
		return true;
	}
#endif
}

// Return current directory
// -> Used for compatibility with QGLE.EXE
bool GLEGetCrDirWin32(string* name) {
#if defined(__WIN32__) || defined(__CYGWIN__)
	TCHAR buffer[1024];
	if (GetCurrentDirectory(1024, buffer) != 0) {
		*name = buffer;
		return true;
	} else {
		return false;
	}
#else
	char buffer[1024];
	if (getcwd(buffer, 1024) == NULL) {
		return false;
	} else {
		*name = buffer;
		return true;
	}
#endif
}

// Change directory
bool GLEChDir(const string& dir) {
#ifdef __WIN32__
	return SetCurrentDirectory(dir.c_str());
#else
	return chdir(dir.c_str()) == 0;
#endif
}

void GLESetGLETop(const string& cmdline) {
	string gle_top = cmdline;
	StripPathComponents(&gle_top, 1);
	// Newer versions have the executable in the "bin" subdir
	if (!GLEFileExists(gle_top + DIR_SEP + "inittex.ini")) {
		StripPathComponents(&gle_top, 1);
	}
	gle_top = "GLE_TOP="+gle_top;
#ifdef __WIN32__
	_putenv(gle_top.c_str());
#endif
}

#ifndef __WIN32__
void GLECloseFDArray(int* fd) {
	for (int i = 0; i < 6; i++) {
		if (fd[i] != -1) close(fd[i]);
		fd[i] = -1;
	}
}
#endif

#define GLEREAD_BUF 10000

#ifdef __WIN32__
void GLEReadFileWin32(HANDLE input, ostream* strm) {
	char buffer[GLEREAD_BUF+1];
	while(true) {
		DWORD nbytes = 0;
		BOOL fSuccess = ReadFile(input, buffer, GLEREAD_BUF, &nbytes, NULL);
		if (!fSuccess || nbytes == 0) {
			CloseHandle(input);
			return;
		} else if (strm != NULL) {
			buffer[nbytes] = 0;
			str_remove_all(buffer, '\r');
			(*strm) << buffer;
		}
	}
}
#else

void GLEReadFile(int file, ostream* strm) {
	char buffer[GLEREAD_BUF+1];
	while(true) {
		int nbytes = read(file, buffer, GLEREAD_BUF);
		if (nbytes <= 0) {
			close(file);
			return;
		} else if (strm != NULL) {
			buffer[nbytes] = 0;
			str_remove_all(buffer, '\r');
			(*strm) << buffer;
		}
	}
}

#endif

int GLERun(const string& cmd, string& result) {
	stringstream strm;
	int res = GLESystem(cmd, true, &strm);
	result = strm.str();
	return res;
}

#define GLESYS_PIPE_RD 0
#define GLESYS_PIPE_WR 1

int GLESystem(const string& cmd, bool wait, stringstream* strm) {
#ifdef __WIN32__
	SECURITY_ATTRIBUTES saAttr;
	saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
	saAttr.bInheritHandle = TRUE;
	saAttr.lpSecurityDescriptor = NULL;
	HANDLE hChildStdinRd, hChildStdinWr, hChildStdoutRd, hChildStdoutWr;
	if (wait) {
		// Create pipe for standard out
		if (!CreatePipe(&hChildStdoutRd, &hChildStdoutWr, &saAttr, 0)) {
			cout << "stdout pipe creation failed" << endl;
			return GLE_SYSTEM_ERROR;
		}
		SetHandleInformation(hChildStdoutRd, HANDLE_FLAG_INHERIT, 0);
		// Create pipe for standard in
		if (!CreatePipe(&hChildStdinRd, &hChildStdinWr, &saAttr, 0)) {
			cout << "stdin pipe creation failed" << endl;
			return GLE_SYSTEM_ERROR;
		}
		SetHandleInformation(hChildStdinWr, HANDLE_FLAG_INHERIT, 0);
	}
	PROCESS_INFORMATION piProcInfo;
	STARTUPINFO siStartInfo;
	ZeroMemory(&piProcInfo, sizeof(PROCESS_INFORMATION));
	ZeroMemory(&siStartInfo, sizeof(STARTUPINFO));
	siStartInfo.cb = sizeof(STARTUPINFO);
	if (wait) {
		siStartInfo.hStdError = hChildStdoutWr;
		siStartInfo.hStdOutput = hChildStdoutWr;
		siStartInfo.hStdInput = hChildStdinRd;
		siStartInfo.dwFlags |= STARTF_USESTDHANDLES;
	}
	BOOL bFuncRetn = CreateProcess(NULL,
	      (CHAR*)cmd.c_str(),                        // command line
	      NULL,                                      // process security attributes
	      NULL,                                      // primary thread security attributes
	      wait,                                      // handles are inherited
	      NORMAL_PRIORITY_CLASS | CREATE_NO_WINDOW,  // creation flags
	      NULL,                                      // use parent's environment
	      NULL,                                      // use parent's current directory
	      &siStartInfo,                              // STARTUPINFO pointer
	      &piProcInfo);                              // receives PROCESS_INFORMATION
	if (bFuncRetn == 0) {
		if (wait) {
			CloseHandle(hChildStdinRd);
			CloseHandle(hChildStdinWr);
			CloseHandle(hChildStdoutRd);
			CloseHandle(hChildStdoutWr);
		}
		return GLE_SYSTEM_ERROR;
	}
	if (wait) {
		CloseHandle(hChildStdinRd);
		CloseHandle(hChildStdinWr);
		CloseHandle(hChildStdoutWr);
		GLEReadFileWin32(hChildStdoutRd, strm);
		WaitForSingleObject(piProcInfo.hProcess, INFINITE);
	}
	CloseHandle(piProcInfo.hProcess);
	CloseHandle(piProcInfo.hThread);
	return GLE_SYSTEM_OK;
#else
	int fd[] = { -1, -1, -1, -1, -1, -1 };
	int* const pin = fd;
	int* const pout = fd+2;
	int* const perr = fd+4;
	if (wait) {
		pipe(pin);
		pipe(pout);
		pipe(perr);
	}
	pid_t pid = fork();
	if (pid == 0) {
		if (*pin >= 0) {
			close(pin[GLESYS_PIPE_WR]);
			dup2(pin[GLESYS_PIPE_RD], STDIN_FILENO);
			close(pin[GLESYS_PIPE_RD]);
		}
		if (*pout >= 0) {
			close(pout[GLESYS_PIPE_RD]);
			dup2(pout[GLESYS_PIPE_WR], STDOUT_FILENO);
			close(pout[GLESYS_PIPE_WR]);
		}
		if (*perr >= 0) {
			close(perr[GLESYS_PIPE_RD]);
			dup2(perr[GLESYS_PIPE_WR], STDERR_FILENO);
			close(perr[GLESYS_PIPE_WR]);
		}
		execl("/bin/sh", "/bin/sh", "-c", cmd.c_str(), NULL);
		_exit(0);
	} else if (pid < 0) {
		GLECloseFDArray(fd);
		return GLE_SYSTEM_ERROR;
	} else if (wait) {
		int status;
		if (*pin >= 0) {
			close(pin[GLESYS_PIPE_RD]);
			close(pin[GLESYS_PIPE_WR]);
		}
		if (*pout >= 0)	{
			close(pout[GLESYS_PIPE_WR]);
		}
		if (*perr >= 0) {
			close(perr[GLESYS_PIPE_WR]);
		}
		GLEReadFile(pout[GLESYS_PIPE_RD], strm);
		GLEReadFile(perr[GLESYS_PIPE_RD], strm);
		// if (strm != NULL) cout << "Result: " << strm->str() << endl;
		waitpid(pid, &status, 0);
	}
#endif
	return GLE_SYSTEM_OK;
}

void GLESleep(int msec) {
#ifdef __WIN32__
	Sleep(msec);
#else
	sleep(msec/1000);
#endif
}

bool GLEFileExists(const string& fname) {
	FILE* f = fopen(fname.c_str(), "rb");
	if (f != NULL) {
		fclose(f);
		return true;
	} else {
		return false;
	}
}

// Create new directory
void MakeDirectory(const string& dir) {
	/* Create directory user read/write/exec, group & others read/exec */
#ifdef __WIN32__
	SECURITY_ATTRIBUTES sec_attr;
	sec_attr.nLength = sizeof(SECURITY_ATTRIBUTES);
	sec_attr.lpSecurityDescriptor = NULL;
	sec_attr.bInheritHandle = FALSE;
	CreateDirectory(dir.c_str(), &sec_attr);
#else
	mkdir(dir.c_str(), S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
#endif
}

// Recursively create directory if it does not yet exist
void EnsureMkDir(const string& dir) {
	if (!IsDirectory(dir)) {
		int done = 0;
		string temp = dir;
		vector<string> comps;
		do {
			string::size_type i = temp.rfind(DIR_SEP);
			if (i != string::npos) {
				comps.push_back(temp.substr(i+1));
				temp = temp.substr(0, i);
			} else {
				comps.push_back(temp);
				done = 1;
			}
		} while (done == 0 && (!IsDirectory(temp)));
		if (done == 0) {
			temp += DIR_SEP;
		} else {
			temp = "";
		}
		for (int i = comps.size()-1; i >= 0; i--) {
			temp += comps[i];
			MakeDirectory(temp);
			if (i > 0) temp += DIR_SEP;
		}
	}
}

// Delete file with extension replaced by given one
bool TryDeleteFile(const string& fname) {
#ifdef __WIN32__
	DeleteFile(fname.c_str());
#else
	unlink(fname.c_str());
#endif
	return true;
}

// Delete file with extension replaced by given one
bool DeleteFileWithExt(const string& fname, const char* ext) {
	string main_name;
	GetMainName(fname, main_name);
	main_name += ext;
	return TryDeleteFile(main_name);
}

// Read line from file
int ReadFileLine(istream& file, string& line) {
    line = "";
    int count = 0;
    char ch = '\n';
    while ((ch == '\n' || ch == '\r') && file.good()) {
        file.read(&ch, 1);
    }
    while (ch != '\n' && ch != '\r' && file.good()) {
        count++;
        line += ch;
        file.read(&ch, 1);
    }
    return count;
}

int ReadFileLineAllowEmpty(istream& file, string& line) {
    line = "";
    char ch;
    int count = 0;
    file.read(&ch, 1);
    while (ch != '\n' && ch != '\r' && !file.eof()) {
        count++;
        line += ch;
        file.read(&ch, 1);
    }
    return count;
}

bool IsDirectory(const string& fname) {
#if defined(__UNIX__) || ( defined(__OS2__) && defined(__EMX__) )
	struct stat stat_buf;
	if (stat((const char *)fname.c_str(), &stat_buf) == 0) {
		return S_ISDIR(stat_buf.st_mode);
	} else {
		return false;
	}
#else
	struct _stat stat_buf;
	if (_stat((const char *)fname.c_str(), &stat_buf) == 0) {
		return (stat_buf.st_mode & _S_IFDIR);
	} else {
		return false;
	}
#endif
}

void GLEFindFiles(const string& dir, const vector<string>& tofind, vector<string*>& result) {
	static int find_files_progress = 0;
	vector<string> subdirs;
	if (find_files_progress++ == 10) {
		cout << "."; fflush(stdout);
		find_files_progress = 0;
	}
#ifdef __WIN32__
	WIN32_FIND_DATA FindFileData;
	string findpattern = dir + DIR_SEP + "*.*";
	HANDLE hFind = FindFirstFile(findpattern.c_str(), &FindFileData);
	if (hFind == INVALID_HANDLE_VALUE) {
		return;
	}
	do {
		const char* name = FindFileData.cFileName;
		if ((FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0) {
			if (!str_i_equals(name, ".") && !str_i_equals(name, "..")) {
				subdirs.push_back(name);
			}
		} else {
			for (vector<string>::size_type i = 0; i < tofind.size(); i++) {
				if (str_i_equals(name, tofind[i].c_str())) {
					string place = dir + DIR_SEP + tofind[i];
					int len_i = (*result[i]).length();
					if (len_i >= 1 && (*result[i])[len_i-1] == ';') {
						if (len_i == 1) (*result[i]) = place + ";";
						else (*result[i]) += place + ";";
					} else {
						(*result[i]) = place;
					}
				}
			}
		}
	} while (FindNextFile(hFind, &FindFileData) != 0);
	FindClose(hFind);
#endif
	for (vector<string>::size_type i = 0; i < subdirs.size(); i++) {
		string nextdir = dir + DIR_SEP + subdirs[i];
		GLEFindFiles(nextdir, tofind, result);
	}
}

string GetActualFilename(string fname) {
	// will return the filename
	// if found in ./ or in any of the include
	// paths, will return empty string if
	// file not found

	std::ifstream file;
	file.open(fname.c_str());
	if (file.is_open()){
		file.close();
		return fname;
	}

	// -- try the IncludePaths
	vector<string> IncludePaths;
	FillIncludePaths(IncludePaths);
	vector<string>::iterator vsi = IncludePaths.begin();
	while(vsi != IncludePaths.end()){
		string p = *vsi+DIR_SEP.c_str()+fname;
		file.open(p.c_str());
		if(file.is_open()){
			file.close();
			return p;
		}
		vsi++;
	}
	return "";
}

#ifdef HAVE_SOCKETS
void GLECloseSocket(SOCKET sock) {
#ifdef __WIN32__
	closesocket(sock);
	WSACleanup();
#endif
#ifdef __UNIX__
	close(sock);
#endif
}
#endif

int GLESendSocket(const string& commands) {
#ifdef HAVE_SOCKETS
#ifdef __WIN32__
	struct WSAData wsaData;
	int nCode = WSAStartup(MAKEWORD(1, 1), &wsaData);
	if (nCode != 0) {
		return -1;
	}
#endif
	struct sockaddr_in name;
	SOCKET sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sock < 0) {
		#ifdef __WIN32__
		WSACleanup();
		#endif
		return -2;
	}
	name.sin_family = AF_INET;
	name.sin_port = htons(6667);
	name.sin_addr.s_addr = inet_addr("127.0.0.1");
	if (connect(sock, (struct sockaddr *) &name, sizeof (name)) < 0) {
		GLECloseSocket(sock);
		return -3;
	}
	int nbsend = send(sock, commands.c_str(), commands.length(), 0);
	if (nbsend != (int)commands.length()) {
		GLECloseSocket(sock);
		return -4;
	}
	GLECloseSocket(sock);
	return 0;
#else
	return -1;
#endif
}

bool GetExeName(const char* appname, string& exe_name) {
#ifdef __WIN32__
	char name[1024];
	DWORD res = GetModuleFileName(NULL, name, 1023);
	if (res > 0) {
		name[res] = 0;
		exe_name = name;

		return true;
	}
#endif
#ifdef __MACOS__
	char name[1024];
	uint32_t path_len = 1023;
	if (_NSGetExecutablePath(name, &path_len) == 0) {
		//str_replace_all(name, "/./", "/");
		exe_name = name;
		str_replace_all(exe_name, "/./", "/");
		return true;
	}
#endif
#if defined(__UNIX__) && !defined(__MACOS__)
	/* try to read location from the /proc/self/exe file */
	char path[PATH_MAX];
	struct stat stat_buf;
	string path2 = "/proc/self/exe";
	while (true) {
		int size = readlink(path2.c_str(), path, PATH_MAX - 1);
		if (size == -1) {
			break;
		}
		path[size] = '\0';
		int i = stat(path, &stat_buf);
		if (i == -1) {
			break;
		}
		if (!S_ISLNK(stat_buf.st_mode)) {
			exe_name = path;
			return true;
		}
		path2 = path;
	}
	/* try to read location from the /proc/self/maps file */
	ifstream maps("/proc/self/maps");
	if (!maps.is_open()) {
		return false;
	}
	string f1 = DIR_SEP + appname;
	string f2 = f1 + ".exe";
	while (!maps.eof()) {
		string line;
		ReadFileLine(maps, line);
		char_separator separator(" ", "");
		tokenizer<char_separator> tokens(line, separator);
		while (tokens.has_more()) {
			exe_name = tokens.next_token();
			if (str_i_ends_with(exe_name, f1.c_str()) || str_i_ends_with(exe_name, f2.c_str())) {
				return true;
			}
		}
	}
	maps.close();
#endif
	return false;
}

StreamTokenizerMax::StreamTokenizerMax(const string& fname, int sep, int max) : m_File(fname.c_str()) {
	m_Sep = sep; m_Max = max; m_IsOK = 1;
	m_LastToken = new char[m_Max+1];
	if (!m_File.is_open()) {
		m_IsOK = 0;
	}
}

StreamTokenizerMax::~StreamTokenizerMax() {
	delete[] m_LastToken;
}

bool StreamTokenizerMax::hasMoreTokens() {
	if (m_IsOK == 1) readNextToken();
	return m_IsOK == 1;
}

const char* StreamTokenizerMax::nextToken() {
	return m_LastToken;
}

void StreamTokenizerMax::close() {
	m_File.close();
}

bool StreamTokenizerMax::isSepChar(char ch) {
	return ch == m_Sep || ch == '\n' || ch == '\r' || ch == 0;
}

void StreamTokenizerMax::readNextToken() {
	char ch = m_Sep;
	while (isSepChar(ch) && !m_File.eof()) {
		m_File.read(&ch, 1);
	}
	int count = 0;
	while (count < m_Max && !isSepChar(ch) && !m_File.eof()) {
		if (ch != m_Sep) m_LastToken[count++] = ch;
		m_File.read(&ch, 1);
	}
	m_LastToken[count] = 0;
	while (!isSepChar(ch) && !m_File.eof()) {
		m_File.read(&ch, 1);
	}
	if (m_File.eof()) {
		m_IsOK = 0;
	}
}

//
// -- the only two dir finding routines needed
//
const int PATH_LENGTH = 256;

void CopyGLETop(char* buf) {
	strcpy(buf, GLE_TOP_DIR.c_str());
}

char *fontdir(char *fname) {
	// simply returns %GLE_TOP%/font
	static char fbuff[PATH_LENGTH];
	CopyGLETop(fbuff);
	strcat(fbuff,DIR_SEP.c_str());
	strcat(fbuff,"font");
	strcat(fbuff,DIR_SEP.c_str());
	strcat(fbuff,fname);
	return fbuff;
}

char *gledir(char *fname) {
	// simply returns %GLE_TOP%
	static char fbuff[PATH_LENGTH];
	CopyGLETop(fbuff);
	strcat(fbuff,DIR_SEP.c_str());
	strcat(fbuff,fname);
	return fbuff;
}
