/*************************************************************************
 *                                                                       *
 * GLE - Graphics Layout Engine <http://www.gle-graphics.org/>           *
 *                                                                       *
 * BSD License                                                           *
 *                                                                       *
 * Copyright (c).                                                        *
 * All rights reserved.                                                  *
 *                                                                       *
 * Redistribution and use in source and binary forms, with or without    *
 * modification, are permitted provided that the following conditions    *
 * are met:                                                              *
 *                                                                       *
 *  * Redistributions of source code must retain the above copyright     *
 *  notice, this list of conditions and the following disclaimer.        *
 *                                                                       *
 *  * Redistributions in binary form must reproduce the above            *
 *  copyright notice, this list of conditions and the following          *
 *  disclaimer in the documentation and/or other materials provided      *
 *  with the distribution.                                               *
 *                                                                       *
 *  * Neither the name of GLE nor the names of its contributors may      *
 *  be used to endorse or promote products derived from this software    *
 *  without specific prior written permission.                           *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT      *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT   *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  *
 *                                                                       *
 *************************************************************************/

#include <fstream>

#include "all.h"
#include "cutils.h"
#include "file_io.h"
#include "SourceLine.h"
#include "core.h"
#include "cmdline.h"
#include "config.h"
#include "core.h"
#include "gleinterface.h"
#include "tokens/stokenizer.h"
#include "texinterface.h"
#include "drawit.h"
#include "mygraph.h"
#include "justify.h"

// TODO make these available through the interface
extern CmdLineObj g_CmdLine;
extern ConfigCollection g_Config;
extern string GLE_WORKING_DIR;
extern string GLE_TOP_DIR;
extern GLEGlobalSource g_GLESource;

void init_option_args(CmdLineObj& cmdline);
GLEScript* load_gle_code_sub(const char* name, CmdLineObj& cmdline) throw(ParserError);
void load_one_file_sub(GLEScript* script, CmdLineObj& cmdline, size_t* exit_code) throw(ParserError);

void GLEScaleSimpleLineProperties(double scale, bool dir, GLEPropertyStore* prop);
void GLEScaleArrowProperties(double scale, bool dir, GLEPropertyStore* prop);

GLEInterface* g_GLEInterface = NULL;

GLEInterface* GLEGetInterfacePointer() {
	if (g_GLEInterface == NULL) {
		GLEGlobalConfig* config = new GLEGlobalConfig();
		config->setCmdLine(&g_CmdLine);
		config->setRCFile(&g_Config);
		g_GLEInterface = new GLEInterface();
		g_GLEInterface->setConfig(config);
	}
	return g_GLEInterface;
}

GLEInterface::GLEInterface() {
	m_Output = new GLEOutputStream();
	m_MakeDrawObjs = false;
	m_CommitMode = false;
	m_FontHash = new StringIntHash();
	m_FontIndexHash = new IntIntHash();

	// Initialize text property store model
	m_TextModel.add(new GLEPropertyFont("Font"));
	GLEPropertyNominal* fontstyle = new GLEPropertyNominal("Font style", GLEPropertyTypeInt, GLEDOPropertyFontStyle);
	fontstyle->addValue("roman", GLEFontStyleRoman);
	fontstyle->addValue("bold", GLEFontStyleBold);
	fontstyle->addValue("italic", GLEFontStyleItalic);
	fontstyle->addValue("bold+italic", GLEFontStyleBoldItalic);
	m_TextModel.add(fontstyle);
	m_TextModel.add(new GLEPropertyHei("Font size"));
	m_TextModel.add(new GLEPropertyColor("Text color"));
	m_TextModel.add(new GLEPropertyJustify("Text justify"));

	m_LineModel.add(new GLEPropertyLWidth("Line width"));
	m_LineModel.add(new GLEPropertyColor("Line color"));
	m_LineModel.add(new GLEProperty("Line style", GLEPropertyTypeInt, GLEDOPropertyLineStyle));
	GLEPropertyNominal* linecap = new GLEPropertyNominal("Line cap", GLEPropertyTypeInt, GLEDOPropertyLineCap);
	linecap->addValue("butt", GLELineCapButt);
	linecap->addValue("round", GLELineCapRound);
	linecap->addValue("square", GLELineCapSquare);
	m_LineModel.add(linecap);

	m_LineModel.add(new GLEPropertyArrowSize("Arrow size"));
	m_LineModel.add(new GLEPropertyArrowAngle("Arrow angle"));
	GLEPropertyNominal* arrowstyle = new GLEPropertyNominal("Arrow style", GLEPropertyTypeInt, GLEDOPropertyArrowStyle);
	arrowstyle->addValue("simple", GLEArrowStyleSimple);
	arrowstyle->addValue("filled", GLEArrowStyleFilled);
	arrowstyle->addValue("empty", GLEArrowStyleEmpty);
	m_LineModel.add(arrowstyle);
	GLEPropertyNominal* arrowtip = new GLEPropertyNominal("Arrow tip", GLEPropertyTypeInt, GLEDOPropertyArrowTip);
	arrowtip->addValue("round", GLEArrowTipRound);
	arrowtip->addValue("sharp", GLEArrowTipSharp);
	m_LineModel.add(arrowtip);

	m_ShapeModel.add(new GLEPropertyLWidth("Line width"));
	m_ShapeModel.add(new GLEPropertyColor("Line color"));
	m_ShapeModel.add(new GLEProperty("Line style", GLEPropertyTypeInt, GLEDOPropertyLineStyle));
	m_ShapeModel.add(new GLEPropertyFillColor("Fill color"));

	m_InitialPS = NULL;
}

GLEInterface::~GLEInterface() {
	delete m_FontHash;
	delete m_FontIndexHash;
}

void GLEInterface::initializeGLE() {
	try {
		// Initialize code in core.cpp
		g_init();
		// Load configuration file
		init_config(&g_Config);
		init_option_args(g_CmdLine);
		do_load_config(g_CmdLine, g_Config);
	} catch (ParserError err) {
		string err_str;
		err.toString(err_str);
		str_uppercase_initial_capital(err_str);
		g_message(err_str);
	}
}

const char* GLEInterface::getGLEVersion() {
	return GLEVN;
}

const char* GLEInterface::getGLETop() {
	return GLE_TOP_DIR.c_str();
}

void GLEInterface::setOutputStream(GLEOutputStream* output) {
	m_Output = output;
}

GLEScript* GLEInterface::newGLEFile(const char* glecode, const char* tmpfile) {
	try {
		string name = tmpfile;
		string in_name = name;
		string dir_name = GLE_WORKING_DIR;
		if (IsAbsPath(in_name)) {
			SplitFileName(in_name, dir_name, in_name);
		}
		GLEChDir(dir_name);
		GLEScript* script = new GLEScript();
		script->setFullName(name);
		script->setFileName(in_name);
		script->setDirName(dir_name);
		g_GLESource.clear();
		GLESourceFile* file = new GLESourceFile();
		file->setFileName(in_name);
		g_GLESource.setMainFile(file);
		char_separator sep("\n");
		tokenizer<char_separator> tokens(glecode, sep);
		while (tokens.has_more()) {
			string line = tokens.next_token();
			str_trim_both(line);
			GLESourceLine* sline = file->addLine();
			sline->setCode(line);
		}
		file->trim(1);
		g_GLESource.initFromMain();
		return script;
	} catch (ParserError err) {
		string err_str;
		err.toString(err_str);
		str_uppercase_initial_capital(err_str);
		g_message(err_str);
		return NULL;
	}
}

GLEScript* GLEInterface::loadGLEFile(const char* glefile) {
	try {
		g_set_compatibility(GLE_COMPAT_MOST_RECENT);
		return load_gle_code_sub(glefile, g_CmdLine);
	} catch (ParserError err) {
		string err_str;
		err.toString(err_str);
		str_uppercase_initial_capital(err_str);
		g_message(err_str);
		return NULL;
	}
}

void GLEInterface::saveGLEFile(GLEScript* script, const char* glefile) {
	ofstream file(glefile);
	GLESourceFile* main = g_GLESource.getMainFile();
	for (int i = 0; i < main->getNbLines(); i++) {
		GLESourceLine* line = main->getLine(i);
		file << line->getPrefix() << line->getCode() << endl;
	}
	file.close();
}

void GLEInterface::showGLEFile(GLEScript* script) {
	cout << "Script:" << endl;
	GLESourceFile* file = g_GLESource.getMainFile();
	for (int i = 0; i < file->getNbLines(); i++) {
		GLESourceLine* line = file->getLine(i);
		cout << line->getCode() << endl;
	}
}

void GLEInterface::setCompatibilityMode(const char* version) {
	CmdLineArgString* arg = (CmdLineArgString*)g_CmdLine.createOption(GLE_OPT_COMPAT)->getArg(0);
	arg->setValue(version);
	g_set_compatibility(version);
}

void GLEInterface::renderGLE(GLEScript* script, const char* outfile, int device) {
	m_Script = script;
	if (script == NULL) {
		cerr << "GLEInterface::renderGLE(): script == NULL" << endl;
		return;
	}
	try {
		size_t exit_code;
		// Set the device
		CmdLineArgSet* dev = (CmdLineArgSet*)g_CmdLine.createOption(GLE_OPT_DEVICE)->getArg(0);
		dev->reset();
		dev->addValue(device);
		// Set the output file name
		CmdLineArgString* o_file = (CmdLineArgString*)g_CmdLine.createOption(GLE_OPT_OUTPUT)->getArg(0);
		o_file->setValue(outfile);
		if (isMakeDrawObjects()) script->clear();
		load_one_file_sub(script, g_CmdLine, &exit_code);
		m_Output->setExitCode(get_nb_errors());
	} catch (ParserError err) {
		string err_str;
		err.toString(err_str);
		str_uppercase_initial_capital(err_str);
		g_message(err_str);
		m_Output->setExitCode(1);
	}
}

void handleNewProperties(GLEPropertyStore* props) {
	vector<GLEProperty*> changed;
	GLEPropertyStoreModel* model = props->getModel();
	for (int i = 0; i < model->getNumberOfProperties(); i++) {
		GLEProperty* prop = model->getProperty(i);
		if (!prop->isEqualToState(props)) {
			prop->updateState(props);
			changed.push_back(prop);
		}
	}
	if (changed.size() != 0) {
		stringstream ss;
		ss << "set";
		for (vector<GLEProperty*>::size_type i = 0; i < changed.size(); i++) {
			GLEProperty* prop = changed[i];
			prop->createSetCommandGLECode(ss, props->getPropertyValue(prop));
		}
		g_GLESource.addLine(ss.str());
	}
}

void GLEInterface::commitChangesGLE(GLEScript* script) {
	m_Script = script;
	if (script == NULL) {
		cerr << "GLEInterface::commitChangesGLE(): script == NULL" << endl;
		return;
	}
	try {
		setCommitMode(true);
		setMakeDrawObjects(true);
		GLEDevice* old_device = g_set_dummy_device();
		string out_name = "DUMMY.EPS";
		TeXInterface* interface = TeXInterface::getInstance();
		interface->initialize(script->getFileName(), out_name);
		interface->reset();
		script->resetObjectIterator();
		DrawIt(out_name, g_GLESource, &g_CmdLine);
		for (int i = 0; i < script->getNumberNewObjects(); i++) {
			string code;
			GLEDrawObject* obj = script->getNewObject(i);
			if (!obj->hasFlag(GDO_FLAG_DELETED)) {
				obj->createGLECode(code);
				/* add line to code */
				handleNewProperties(obj->getProperties());
				g_GLESource.addLine(code);
				/* compile and execute line of code */
				obj->updateBoundingBox();
				script->addObject(obj);
			}
		}
		g_GLESource.performUpdates();
		script->clearNewObjects();
		script->removeDeletedObjects();
		interface->tryCreateHash();
		g_restore_device(old_device);
		setMakeDrawObjects(false);
		setCommitMode(false);
	} catch (ParserError err) {
		string err_str;
		err.toString(err_str);
		str_uppercase_initial_capital(err_str);
		g_message(err_str);
		m_Output->setExitCode(1);
	}
}

CmdLineObj* GLEInterface::getCmdLine() {
	return m_Config->getCmdLine();
}

int GLEInterface::getNumberOfColors() {
	return GLEGetColorList()->getNbColors();
}

GLEColor* GLEInterface::getColor(int i) {
	return GLEGetColorList()->getColor(i);
}

void GLEInterface::addFont(GLEFont* font) {
	m_FontHash->add_item(font->getName(), m_Fonts.size());
	m_FontIndexHash->add_item(font->getIndex(), m_Fonts.size());
	font->setNumber(m_Fonts.size());
	m_Fonts.push_back(font);
}

// Return the number of available fonts
int GLEInterface::getNumberOfFonts() {
	return m_Fonts.size();
}

// Return the i-the font
GLEFont* GLEInterface::getFont(int i) {
	if (i < 0 || i >= (int)m_Fonts.size()) return m_Fonts[0];
	return m_Fonts[i];
}

// Return the font matching a given name
GLEFont* GLEInterface::getFont(const string& name) {
	int idx = m_FontHash->try_get(name);
	if (idx == -1) return NULL;
	else return m_Fonts[idx];
}

GLEFont* GLEInterface::getFont(const char* name) {
	return getFont(string(name));
}

int GLEInterface::getFontIndex(int font) {
	return m_FontIndexHash->try_get(font);
}

const char* GLEInterface::getInitialPostScript() {
	if (m_InitialPS == NULL) {
		g_select_device(GLE_DEVICE_EPS);
		PSGLEDevice* dev = (PSGLEDevice*)g_get_device_ptr();
		dev->startRecording();
		g_clear();
		dev->startRecording();
		dev->initialPS();
		m_InitialPS = new string();
		dev->getRecordedPostScript(m_InitialPS);
	}
	return m_InitialPS->c_str();
}

const char* GLEInterface::getTerminatePostScript() {
	return "showpage\ngrestore\n";
}

void GLEInterface::renderText(GLETextDO* text, GLEPropertyStore* prop) {
	g_select_device(GLE_DEVICE_EPS);
	PSGLEDevice* dev = (PSGLEDevice*)g_get_device_ptr();
	dev->startRecording();
	g_clear();
	g_resetfont();
	g_scale(72.0/CM_PER_INCH,72.0/CM_PER_INCH);
	g_translate(1.0*CM_PER_INCH/72, 1.0*CM_PER_INCH/72);
	dev->startRecording();
	// Select the right color
	GLEColor* color = prop->getColorProperty(GLEDOPropertyColor);
	g_set_color(color);
	// Select the right size
	double hei = prop->getRealProperty(GLEDOPropertyFontSize);
	g_set_hei(hei);
	g_set_font_width(-1);
	g_set_line_style("1");
	g_set_line_width(0.02);
	// Select the font
	GLEFont* font = prop->getFontProperty(GLEDOPropertyFont);
	if (font == NULL) font = getFont("rm");
	if (font != NULL) {
		GLEFontStyle style = (GLEFontStyle)prop->getIntProperty(GLEDOPropertyFontStyle);
		if (font->hasStyle(style)) font = font->getStyle(style);
		g_set_font(font->getIndex());
	}
	char* str = (char*)text->getTextC();
	// Calculate the text's size
	double x1, x2, y1, y2;
	g_measure(str, &x1, &x2, &y2, &y1);
	text->initBB(x2-x1, y2-y1, -y1);
// DRAW BOUNDING BOX
/*	g_set_color(0X01FF0000);
	g_set_line_width(0.005);
	g_move(0.0,0.0);
	g_line(x2-x1, 0.0);
	g_line(x2-x1, y2-y1);
	g_line(0.0, y2-y1);
	g_line(0.0, 0.0);
	dev->flush();
	g_set_color(0X01000000);*/
	// Draw the string
	g_move(0.0, 0.0);
	g_jtext(JUST_BL);
	// Return the resulting PostScript code
	dev->getRecordedPostScript(text->getPostScriptPtr());
/*
	string code;
	code += getInitialPostScript();
	code += (*text->getPostScriptPtr());
	code += getTerminatePostScript();
	cout << "PS:" << endl << code << endl; */
}

GLETextDO* GLEInterface::renderText(const char* str, GLEPropertyStore* prop) {
	GLETextDO* text = new GLETextDO(0.0, 0.0, str);
	renderText(text, prop);
	return text;
}

const char* GLEInterface::getDeviceFilenameExtension(int device) {
	switch (device) {
		case GLE_DEVICE_PDF:  return "pdf";
		case GLE_DEVICE_PS:   return "ps";
		case GLE_DEVICE_EPS:  return "eps";
		case GLE_DEVICE_PNG:  return "png";
		case GLE_DEVICE_JPEG: return "jpg";
	}
	return "unk";
}

void GLEInterface::clearAllCmdLine() {
	getCmdLine()->clearAll();
}

void GLEInterface::setCmdLineOption(const char* name) {
	getCmdLine()->setHasOption(string(name));
}

void GLEInterface::setCmdLineOptionString(const char* name, const char* value, int arg) {
	getCmdLine()->setOptionString(string(name), string(value), arg);
}

void GLEInterface::getGhostScriptLocation(string* gsloc) {
	*gsloc = "?";
#ifdef __WIN32__
	ConfigSection* tools = g_Config.getSection(GLE_CONFIG_TOOLS);
	string gs_cmd = ((CmdLineArgString*)tools->getOptionValue(GLE_TOOL_GHOSTSCRIPT_CMD))->getValue();
	StripPathComponents(&gs_cmd, 1);
	AddDirSep(gs_cmd);
	gs_cmd += "gsdll32.dll";
	*gsloc = gs_cmd;
#endif
}

GLEErrorMessage::GLEErrorMessage() {
	m_Line = m_Column = -1;
	m_Delta = 0;
}

GLEErrorMessage::~GLEErrorMessage() {
}

GLEOutputStream::GLEOutputStream() {
	m_ExitCode = 0;
}

GLEOutputStream::~GLEOutputStream() {
}

void GLEOutputStream::println() {
	cout << endl;
}

void GLEOutputStream::println(const char* str) {
	cout << str << endl;
}

void GLEOutputStream::error(GLEErrorMessage* msg) {
	const char* file = msg->getFile();
	const char* abbrev = msg->getLineAbbrev();
	stringstream output;
	output << endl;
	output << ">> " << file << " (" << msg->getLine() << ")";
	if (abbrev[0] != 0) {
		output << " |" << abbrev << "|";
	}
	if (msg->getColumn() != -1) {
		char number[50];
		output << endl;
		output << ">> ";
		sprintf(number, "%d", msg->getLine());
		int nbspc = strlen(file) + strlen(number) + 4 + msg->getColumn() - msg->getDelta();
		for (int i = 0; i < nbspc; i++) {
			output << " ";
		}
		output << "^";
	}
	output << msg->getErrorMsg();
	g_message(output.str().c_str());
}

GLEDrawObject::GLEDrawObject() {
	m_Flag = 0;
	m_Properties = NULL;
}

GLEDrawObject::~GLEDrawObject() {
	if (m_Properties != NULL) delete m_Properties;
}

void GLEDrawObject::initProperties(GLEInterface* iface) {
}

void GLEDrawObject::createGLECode(string& code) {
	code = "";
}

void GLEDrawObject::getBoundingBox(GLERectangle* box) {
}

void GLEDrawObject::getPSBoundingBox(GLERectangle* box) {
}

const char* GLEDrawObject::getPostScriptCode() {
	return "";
}

void GLEDrawObject::updateBoundingBox() {
}

void GLEDrawObject::applyTransformation(bool dir) {
}

void GLEDrawObject::applyTransformationPt(GLEPoint* pt, bool dir) {
	if (dir) {
		double devx, devy;
		g_dev(pt->getX(), pt->getY(), &devx, &devy);
		pt->setXY((devx-1.0)/72.0*CM_PER_INCH, (devy-1.0)/72.0*CM_PER_INCH);
	} else {
		double xp, yp;
		g_undev(pt->getX()*72.0/CM_PER_INCH+1.0, pt->getY()*72.0/CM_PER_INCH+1.0, &xp, &yp);
		pt->setXY(xp, yp);
	}
}

GLEDrawObject* GLEDrawObject::clone() {
	return NULL;
}

GLEDrawObject* GLEDrawObject::deepClone() {
	GLEDrawObject* res = clone();
	GLEPropertyStore* store = getProperties();
	if (store != NULL) {
		res->setProperties(store->clone());
	}
	return res;
}

bool GLEDrawObject::approx(GLEDrawObject* other) {
	return false;
}

void GLEDrawObject::draw() {
}

void GLEDrawObject::setProperties(GLEPropertyStore* store) {
	if (m_Properties != NULL) delete m_Properties;
	m_Properties = store;
}

GLEScript::GLEScript() {
}

GLEDrawObject* GLEScript::newGLEObject(GLEDrawObjectType type) {
	GLEDrawObject* obj = NULL;
	switch (type) {
		case GDOLine:
			obj = new GLELineDO();
			obj->initProperties(GLEGetInterfacePointer());
			break;
		case GDOEllipse:
			obj = new GLEEllipseDO();
			obj->initProperties(GLEGetInterfacePointer());
			break;
		case GDOArc:
			obj = new GLEArcDO();
			obj->initProperties(GLEGetInterfacePointer());
			break;
		case GDOText:
			obj = new GLETextDO();
			obj->initProperties(GLEGetInterfacePointer());
			break;
		default:
			break;
	}
	m_NewObjs.push_back(obj);
	return obj;
}

void GLEScript::cancelObject(GLEDrawObject* obj) {
	int pos = m_NewObjs.size()-1;
	while (pos > 0) {
		if (m_NewObjs[pos] == obj) {
			m_NewObjs.erase(m_NewObjs.begin()+pos);
			break;
		}
		pos--;
	}
	delete obj;
}

GLEScript::~GLEScript() {
}

GLEDrawObject* GLEScript::nextObject() {
	if (m_CurrObject < getNumberObjects()) {
		return getObject(m_CurrObject++);
	} else {
		return NULL;
	}
}

void GLEScript::clearNewObjects() {
//	for (vector<GLEDrawObject*>::size_type i = 0; i < m_NewObjs.size(); i++) {
//		delete m_NewObjs[i];
//	}
	m_NewObjs.clear();
}

GLEDrawObjectType GLEScript::getType() {
	return GDOScript;
}

GLELineDO::GLELineDO() {
	m_Arrow = GLEHasArrowNone;
}

GLELineDO::GLELineDO(const GLEPoint& p1, const GLEPoint& p2) : m_P1(p1), m_P2(p2) {
	m_Arrow = GLEHasArrowNone;
}

GLELineDO::GLELineDO(double x1, double y1, double x2, double y2) : m_P1(x1,y1), m_P2(x2,y2) {
	m_Arrow = GLEHasArrowNone;
}

GLELineDO::~GLELineDO() {
}

GLEDrawObjectType GLELineDO::getType() {
	return GDOLine;
}

void GLELineDO::initProperties(GLEInterface* iface) {
	m_Properties = new GLEPropertyStore(iface->getLinePropertyStoreModel());
	GLEInitLineProperties(m_Properties);
	GLEInitArrowProperties(m_Properties);
}

void GLELineDO::createGLECode(string& code) {
	double ox, oy;
	g_get_xy(&ox,&oy);
	stringstream str;
// FIXME: if we need this implementation -> need to detect amove at start of command
//        if such amove is there -> need to include it again!!
//        only don't include amove if position is correct from _previous_ line
/*	if (getP1().approx(ox, oy)) {
		str << "aline " << getP2().getX() << " " << getP2().getY();
	} else if (getP2().approx(ox, oy)) {
		str << "aline " << getP1().getX() << " " << getP1().getY();
		m_P1.swap(m_P2);
	} else {*/
		str << "amove " << getP1().getX() << " " << getP1().getY() << "; ";
		str << "aline " << getP2().getX() << " " << getP2().getY();
/*	}*/
	GLEHasArrow arrow = getArrow();
	if (arrow == (GLEHasArrowStart | GLEHasArrowEnd)) {
		str << " arrow both";
	} else if (arrow == GLEHasArrowStart) {
		str << " arrow start";
	} else if (arrow == GLEHasArrowEnd) {
		str << " arrow end";
	}
	code = str.str();
}

void GLELineDO::updateBoundingBox() {
	g_set_bounds(getP1().getX(), getP1().getY());
	g_set_bounds(getP2().getX(), getP2().getY());
	g_move(getP2().getX(), getP2().getY());
}

void GLELineDO::applyTransformation(bool dir) {
	applyTransformationPt(&m_P1, dir);
	applyTransformationPt(&m_P2, dir);
	double scale = g_get_avg_scale();
	GLEScaleSimpleLineProperties(scale, dir, getProperties());
	GLEScaleArrowProperties(scale, dir, getProperties());
}

GLEDrawObject* GLELineDO::clone() {
	GLELineDO* cl = new GLELineDO(getP1(), getP2());
	cl->setArrow(getArrow());
	return cl;
}

bool GLELineDO::approx(GLEDrawObject* other) {
	GLELineDO* otherc = (GLELineDO*)other;
	return getP1().approx(otherc->getP1()) && getP2().approx(otherc->getP2()) && getArrow() == otherc->getArrow();
}

GLEEllipseDO::GLEEllipseDO() {
	m_Rx = m_Ry = 0.0;
}

GLEEllipseDO::GLEEllipseDO(double x0, double y0, double rx, double ry) : m_Center(x0, y0) {
	m_Rx = rx;
	m_Ry = ry;
}

GLEEllipseDO::GLEEllipseDO(const GLEPoint& c, double rx, double ry) : m_Center(c) {
	m_Rx = rx;
	m_Ry = ry;
}

GLEEllipseDO::GLEEllipseDO(double x0, double y0, double r) : m_Center(x0, y0) {
	m_Rx = m_Ry = r;
}

GLEEllipseDO::~GLEEllipseDO() {
}

GLEDrawObjectType GLEEllipseDO::getType() {
	return GDOEllipse;
}

void GLEEllipseDO::setRadius(double r) {
	m_Rx = m_Ry = r;
}

GLEPoint GLEEllipseDO::getPoint(GLEJustify just) {
	switch (just) {
		case GLEJustifyRC:
			return GLEPoint(m_Center.getX()+m_Rx, m_Center.getY());
		case GLEJustifyTL:
			return GLEPoint(m_Center.getX()-m_Rx, m_Center.getY()+m_Ry);
		case GLEJustifyBR:
			return GLEPoint(m_Center.getX()+m_Rx, m_Center.getY()-m_Ry);
		default:
			break;
	}
	return GLEPoint();
}

void GLEEllipseDO::initProperties(GLEInterface* iface) {
	m_Properties = new GLEPropertyStore(iface->getShapePropertyStoreModel());
	GLEInitSimpleLineProperties(m_Properties);

	/* move to init shape fill color */
	rgb01 rgb;
	colortyp color;
	g_get_fill((int*)&color);
	GLEColor* gcolor = new GLEColor();
	if (color.l == (int)GLE_FILL_CLEAR) {
		gcolor->setTransparent(true);
	} else {
		g_colortyp_to_rgb01(&color, &rgb);
		gcolor->setRGB(rgb.red, rgb.green, rgb.blue);
	}
	m_Properties->setColorProperty(GLEDOPropertyFillColor, gcolor);
}

void GLEEllipseDO::createGLECode(string& code) {
	double ox, oy;
	g_get_xy(&ox,&oy);
	stringstream str;
//	if (!getCenter().approx(ox, oy)) {
		str << "amove " << getCenter().getX() << " " << getCenter().getY() << "; ";
//	}
	if (isCircle()) {
		str << "circle " << m_Rx;
	} else {
		str << "ellipse " << m_Rx << " " << m_Ry;
	}
	code = str.str();
}

void GLEEllipseDO::updateBoundingBox() {
	g_move(getCenter().getX(), getCenter().getY());
	g_set_bounds(getCenter().getX()-m_Rx,getCenter().getY()-m_Ry);
	g_set_bounds(getCenter().getX()+m_Rx,getCenter().getY()+m_Ry);
}

void GLEEllipseDO::applyTransformation(bool dir) {
	applyTransformationPt(&m_Center, dir);
	double sx, sy;
	g_get_scale(&sx, &sy);
	if (dir) {
		m_Rx *= sx;
		m_Ry *= sy;
	} else {
		m_Rx /= sx;
		m_Ry /= sy;
	}
	double scale = (sx+sy)/2.0;
	GLEScaleSimpleLineProperties(scale, dir, getProperties());
}

GLEDrawObject* GLEEllipseDO::clone() {
	return new GLEEllipseDO(m_Center, m_Rx, m_Ry);
}

bool GLEEllipseDO::approx(GLEDrawObject* other) {
	GLEEllipseDO* otherc = (GLEEllipseDO*)other;
	return m_Center.approx(otherc->getCenter()) &&
	       fabs(m_Rx - otherc->getRadiusX()) < 1e-6 &&
	       fabs(m_Ry - otherc->getRadiusY()) < 1e-6;
}

GLEArcDO::GLEArcDO() {
	m_Angle1 = m_Angle2 = 0.0;
}

GLEArcDO::GLEArcDO(double x0, double y0, double r, double a1, double a2) : GLEEllipseDO(x0, y0, r) {
	m_Angle1 = a1; m_Angle2 = a2;
}

GLEArcDO::GLEArcDO(double x0, double y0, double rx, double ry, double a1, double a2) : GLEEllipseDO(x0, y0, rx, ry) {
	m_Angle1 = a1; m_Angle2 = a2;
}

GLEArcDO::~GLEArcDO() {
}

GLEDrawObjectType GLEArcDO::getType() {
	return GDOArc;
}

void GLEArcDO::initProperties(GLEInterface* iface) {
	m_Properties = new GLEPropertyStore(iface->getLinePropertyStoreModel());
	GLEInitLineProperties(m_Properties);
	GLEInitArrowProperties(m_Properties);
}

void GLEArcDO::createGLECode(string& code) {
	double ox, oy;
	g_get_xy(&ox,&oy);
	stringstream str;
//	if (!getCenter().approx(ox, oy)) {
		str << "amove " << getCenter().getX() << " " << getCenter().getY() << "; ";
//	}
	double angle2 = getNormalizedAngle2();
	if (isCircle()) {
		str << "arc " << m_Rx << " " << m_Angle1 << " " << angle2;
	} else {

		str << "elliptical_arc " << m_Rx << " " << m_Ry << " " << m_Angle1 << " " << angle2;
	}
	code = str.str();
}

void GLEArcDO::updateBoundingBox() {
	g_move(m_Center.getX(), m_Center.getY());
}

GLEDrawObject* GLEArcDO::clone() {
	return new GLEArcDO(m_Center.getX(), m_Center.getY(), m_Rx, m_Ry, m_Angle1, m_Angle2);
}

bool GLEArcDO::approx(GLEDrawObject* other) {
	GLEArcDO* otherc = (GLEArcDO*)other;
	return GLEEllipseDO::approx(other) &&
	       fabs(m_Angle1 - otherc->getAngle1()) < 1e-6 &&
	       fabs(m_Angle2 - otherc->getAngle2()) < 1e-6;
}

void GLEArcDO::normalize() {
	m_Angle2 = getNormalizedAngle2();
}

double GLEArcDO::getNormalizedAngle2() {
	if (m_Angle2 < m_Angle1) {
		// Correctly normalize a2
		return m_Angle2 + 360 * ceil((m_Angle1 - m_Angle2)/360.0);
	} else {
		return m_Angle2;
	}
}


GLEPoint& GLEArcDO::getPoint1(GLEPoint& pt) {
	pt.set(m_Center);
	pt.add(m_Rx*cos(m_Angle1*GLE_PI/180.0), m_Ry*sin(m_Angle1*GLE_PI/180.0));
	return pt;
}

GLEPoint& GLEArcDO::getPoint2(GLEPoint& pt) {
	pt.set(m_Center);
	pt.add(m_Rx*cos(m_Angle2*GLE_PI/180.0), m_Ry*sin(m_Angle2*GLE_PI/180.0));
	return pt;
}

GLEPoint& GLEArcDO::getPointMid(GLEPoint& pt) {
	pt.set(m_Center);
	double angle2 = getNormalizedAngle2();
	double angleMid = (m_Angle1 + angle2)/2;
	pt.add(m_Rx*cos(angleMid*GLE_PI/180.0), m_Ry*sin(angleMid*GLE_PI/180.0));
	return pt;
}

GLETextDO::GLETextDO() {
}

GLETextDO::GLETextDO(GLEPoint& position, const string& text) : m_Position(position), m_Text(text) {
}

GLETextDO::GLETextDO(double xp, double yp, const string& text) : m_Position(xp, yp), m_Text(text) {
}

GLETextDO::~GLETextDO() {
}

GLEDrawObjectType GLETextDO::getType() {
	return GDOText;
}

const char* GLETextDO::getPostScriptCode() {
	return m_PostScript.c_str();
}

void GLETextDO::getPSBoundingBox(GLERectangle* box) {
	box->copy(&m_PSBoundingBox);
}

void GLETextDO::initBB(double width, double height, double baseline) {
	m_PSBoundingBox.setXMin(0.0);
	m_PSBoundingBox.setXMax(width);
	m_PSBoundingBox.setYMin(0.0);
	m_PSBoundingBox.setYMax(height);
	m_BaseLine = baseline;
}

void GLETextDO::createGLECode(string& code) {
	double ox, oy;
	g_get_xy(&ox,&oy);
	stringstream str;
//	if (!getPosition().approx(ox, oy)) {
		str << "amove " << getPosition().getX() << " " << getPosition().getY() << "; ";
//	}
	str << "write \"" << getText() << "\"";
	code = str.str();
}

void GLETextDO::updateBoundingBox() {
}

void GLETextDO::applyTransformation(bool dir) {
	applyTransformationPt(&m_Position, dir);
	GLEPropertyStore* props = getProperties();
	if (props != NULL) {
		double scale = g_get_avg_scale();
		if (scale > 0) {
			double hei = props->getRealProperty(GLEDOPropertyFontSize);
			if (dir) {
				hei *= scale;
			} else {
				hei /= scale;
			}
			props->setRealProperty(GLEDOPropertyFontSize, hei);
		}
	}
}

GLEDrawObject* GLETextDO::clone() {
	return new GLETextDO(getPosition(), getText());
}

bool GLETextDO::approx(GLEDrawObject* other) {
	GLETextDO* othert = (GLETextDO*)other;
	return m_Position.approx(othert->getPosition()) &&
	       m_Text == othert->getText();
}

void GLETextDO::initProperties(GLEInterface* iface) {
	int justify, font;
	double fontsize;
	m_Properties = new GLEPropertyStore(iface->getTextPropertyStoreModel());
	GLEInitColorProperty(m_Properties);
	g_get_hei(&fontsize);
	m_Properties->setRealProperty(GLEDOPropertyFontSize, fontsize);
	g_get_just(&justify);
	m_Properties->setIntProperty(GLEDOPropertyJustify, justify);
	g_get_font(&font);
	// this lookup should be removed if "core.cpp" keeps pointer to GLEFont
	font = iface->getFontIndex(font);
	if (font == -1) font = 0;
	m_Properties->setFontProperty(GLEDOPropertyFont, iface->getFont(font));
}

void GLETextDO::render(GLEInterface* iface) {
	iface->renderText(this, getProperties());
}

GLEComposedObject::GLEComposedObject() {
}

GLEComposedObject::~GLEComposedObject() {
}

void GLEComposedObject::clear() {
//	for (vector<GLEDrawObject*>::size_type i = 0; i < m_Objs.size(); i++) {
//		delete m_Objs[i];
//	}
	m_Objs.clear();
}

void GLEComposedObject::removeDeletedObjects() {
	int nbDeleted = 0;
	int nbTotal = getNumberObjects();
	for (int i = 0; i < nbTotal; i++) {
		GLEDrawObject* obj = getObject(i);
		if (obj->hasFlag(GDO_FLAG_DELETED)) {
			nbTotal--;
			nbDeleted++;
		}
//		delete obj;
		setObject(i, getObject(i+nbDeleted));
	}
	setNumberObjects(nbTotal);
}

GLEFont::GLEFont() {
	m_Bold = m_Italic = m_BoldItalic = NULL;
	m_Index = m_Number = 0;
}

GLEFont::~GLEFont() {
}

GLEFont* GLEFont::getStyle(GLEFontStyle style) {
	switch (style) {
		case GLEFontStyleRoman:
			return this;
		case GLEFontStyleBold:
			return m_Bold;
		case GLEFontStyleItalic:
			return m_Italic;
		case GLEFontStyleBoldItalic:
			return m_BoldItalic;
		default:
			return NULL;
	}
}

void GLEFont::setStyle(GLEFontStyle style, GLEFont* font) {
	switch (style) {
		case GLEFontStyleBold:
			m_Bold = font;
			break;
		case GLEFontStyleItalic:
			m_Italic = font;
			break;
		case GLEFontStyleBoldItalic:
			m_BoldItalic = font;
			break;
		default:
			break;
	}
}

GLEColor::GLEColor() {
	m_Red = m_Blue = m_Green = 0.0;
	m_Transparent = true;
	m_Name = NULL;
}

GLEColor::GLEColor(double r, double g, double b) {
	m_Red = r;
	m_Green = g;
	m_Blue = b;
	m_Transparent = false;
	m_Name = NULL;
}

GLEColor::~GLEColor() {
	if (m_Name != NULL) delete m_Name;
}

void GLEColor::toString(ostream& out) {
	char col[80];
	if (isTransparent()) {
		out << "clear";
	} else {
		out << "#";
		sprintf(col, "%.2X", getRedI());
		out << col;
		sprintf(col, "%.2X", getGreenI());
		out << col;
		sprintf(col, "%.2X", getBlueI());
		out << col;
	}
}

const char* GLEColor::getName() {
	return m_Name != NULL ? m_Name->c_str() : "";
}

void GLEColor::setName(const string& name) {
	if (m_Name != NULL) delete m_Name;
	m_Name = new string(name);
}

void GLEColor::setRGB(double r, double g, double b) {
	m_Red = r; m_Green = g; m_Blue = b;
	m_Transparent = false;
}

void GLEColor::setRGB255(int r, int g, int b) {
	m_Red   = ((double)r) / 255.0;
	m_Green = ((double)g) / 255.0;
	m_Blue  = ((double)b) / 255.0;
	m_Transparent = false;
}

void GLEColor::setHexValue(unsigned int v) {
	unsigned int red   = (v >> 16) & 0xFF;
	unsigned int green = (v >> 8) & 0xFF;
	unsigned int blue  = v & 0xFF;
	m_Red   = ((double)red) / 255.0;
	m_Green = ((double)green) / 255.0;
	m_Blue  = ((double)blue) / 255.0;
	// cout << "RGB(" << red << "," << green << "," << blue << ") -> ";
	// cout << "RGB(" << m_Red << "," << m_Green << "," << m_Blue << ")" << endl;
	m_Transparent = false;
}

unsigned int GLEColor::getHexValueGLE() {
	if (isTransparent()) {
		return GLE_FILL_CLEAR;
	} else {
		unsigned int red = float_to_color_comp(m_Red);
		unsigned int green = float_to_color_comp(m_Green);
		unsigned int blue = float_to_color_comp(m_Blue);
		return 0x01000000 | (red << 16) | (green << 8) | blue;
	}
}

GLEPropertyStore::GLEPropertyStore(GLEPropertyStoreModel* model) {
	m_Model = model;
	m_Values.resize(model->getNumberOfProperties());
}

GLEPropertyStore::~GLEPropertyStore() {
}

GLEPropertyStore* GLEPropertyStore::clone() {
	GLEPropertyStore* result = new GLEPropertyStore(getModel());
	for (size_t i = 0; i < m_Values.size(); i++) {
		result->setPropertyValue(i, m_Values[i]);
	}
	return result;
}

void GLEPropertyStore::getPropertyAsString(GLEPropertyID id, string* result) {
	int idx = m_Model->find(id);
	GLEProperty* prop = m_Model->getProperty(idx);
	prop->getPropertyAsString(result, m_Values[idx]);
}

GLEPropertyStoreModel::GLEPropertyStoreModel() {
	m_Hash = new IntIntHash();
}

GLEPropertyStoreModel::~GLEPropertyStoreModel() {
	delete m_Hash;
}

int GLEPropertyStoreModel::find(GLEPropertyID id) {
	return m_Hash->try_get(id);
}

void GLEPropertyStoreModel::add(GLEProperty* prop) {
	int id = m_Properties.size();
	m_Properties.push_back(prop);
	prop->setIndex(id);
	m_Hash->add_item(prop->getID(), id);
}

GLEProperty::GLEProperty(const char* name, const char* cmdname, GLEPropertyType type, GLEPropertyID id) {
	m_Name = name;
	m_Type = type;
	m_ID = id;
	m_SetCmdName = cmdname;
	m_Index = -1;
}

GLEProperty::GLEProperty(const char* name, GLEPropertyType type, GLEPropertyID id) {
	m_Name = name;
	m_Type = type;
	m_ID = id;
	m_SetCmdName = NULL;
	m_Index = -1;
}

GLEProperty::~GLEProperty() {
}

void GLEProperty::getPropertyAsString(string* result, GLEPropertyValue& value) {
	GLEColor* color = NULL;
	GLEFont* font = NULL;
	stringstream str;
	switch (m_Type) {
		case GLEPropertyTypeInt:
			str << value.IntValue;
			break;
		case GLEPropertyTypeBool:
			if (value.BoolValue) str << "yes";
			else str << "no";
			break;
		case GLEPropertyTypeReal:
			str << value.RealValue;
			break;
		case GLEPropertyTypeString:
			break;
		case GLEPropertyTypeColor:
			color = (GLEColor*)value.Pointer;
			color->toString(str);
			break;
		case GLEPropertyTypeFont:
			font = (GLEFont*)value.Pointer;
			str << font->getName();
			break;
	}
	*result = str.str();
}

void GLEProperty::createSetCommandGLECode(ostream& os, GLEPropertyValue& value) {
	if (getSetCommandName() != NULL) {
		string str;
		getPropertyAsString(&str, value);
		os << " " << getSetCommandName() << " " << str;
	}
}

bool GLEProperty::isEqualToState(GLEPropertyStore* store) {
	return true;
}

void GLEProperty::updateState(GLEPropertyStore* store) {
}

GLEPropertyNominal::GLEPropertyNominal(const char* name, GLEPropertyType type, GLEPropertyID id) : GLEProperty(name, type, id) {
	m_Value2Name = new IntIntHash();
	m_Name2Value = new StringIntHash();
}

GLEPropertyNominal::~GLEPropertyNominal() {

}

void GLEPropertyNominal::addValue(const char* name, int value) {
	int pos = m_NomValues.size();
	m_NomValues.push_back(name);
	m_Value2Name->add_item(value, pos);
	m_Name2Value->add_item(name, value);
}

void GLEPropertyNominal::getPropertyAsString(string* result, GLEPropertyValue& value) {
	int pos = m_Value2Name->try_get(value.IntValue);
	if (pos == -1) {
		stringstream res;
		res << value.IntValue;
		*result = res.str();
	} else {
		*result = m_NomValues[pos];
	}
}

GLEPropertyLWidth::GLEPropertyLWidth(const char* name) : GLEProperty(name, "lwidth", GLEPropertyTypeReal, GLEDOPropertyLineWidth) {
}

GLEPropertyLWidth::~GLEPropertyLWidth() {
}

bool GLEPropertyLWidth::isEqualToState(GLEPropertyStore* store) {
	double lwidth;
	g_get_line_width(&lwidth);
	return fabs(lwidth-store->getRealProperty(this)) < 1e-18;
}

void GLEPropertyLWidth::updateState(GLEPropertyStore* store) {
	g_set_line_width(store->getRealProperty(this));
}

GLEPropertyColor::GLEPropertyColor(const char* name) : GLEProperty(name, "color", GLEPropertyTypeColor, GLEDOPropertyColor) {
}

GLEPropertyColor::~GLEPropertyColor() {
}

bool GLEPropertyColor::isEqualToState(GLEPropertyStore* store) {
	rgb01 rgb;
	colortyp color;
	g_get_colortyp(&color);
	g_colortyp_to_rgb01(&color, &rgb);
	GLEColor* gle_color = store->getColorProperty(this);
	return fabs(gle_color->getRed()-rgb.red) < 1e-18 &&
	       fabs(gle_color->getBlue()-rgb.blue) < 1e-18 &&
	       fabs(gle_color->getGreen()-rgb.green) < 1e-18;
}

void GLEPropertyColor::updateState(GLEPropertyStore* store) {
	g_set_color(store->getColorProperty(this));
}

GLEPropertyFillColor::GLEPropertyFillColor(const char* name) : GLEProperty(name, "fill", GLEPropertyTypeColor, GLEDOPropertyFillColor) {
}

GLEPropertyFillColor::~GLEPropertyFillColor() {
}

bool GLEPropertyFillColor::isEqualToState(GLEPropertyStore* store) {
	rgb01 rgb;
	colortyp color;
	g_get_fill_colortyp(&color);
	GLEColor* gle_color = store->getColorProperty(this);
	if (gle_color->isTransparent() != (color.l == (int)GLE_FILL_CLEAR)) {
		return false;
	}
	g_colortyp_to_rgb01(&color, &rgb);
	return fabs(gle_color->getRed()-rgb.red) < 1e-18 &&
	       fabs(gle_color->getBlue()-rgb.blue) < 1e-18 &&
	       fabs(gle_color->getGreen()-rgb.green) < 1e-18;
}

void GLEPropertyFillColor::updateState(GLEPropertyStore* store) {
	g_set_fill(store->getColorProperty(this));
}

GLEPropertyHei::GLEPropertyHei(const char* name) : GLEProperty(name, "hei", GLEPropertyTypeReal, GLEDOPropertyFontSize) {
}

GLEPropertyHei::~GLEPropertyHei() {
}

bool GLEPropertyHei::isEqualToState(GLEPropertyStore* store) {
	double fontsize;
	g_get_hei(&fontsize);
	return fabs(store->getRealProperty(this)-fontsize) < 1e-18;
}

void GLEPropertyHei::updateState(GLEPropertyStore* store) {
	g_set_hei(store->getRealProperty(this));
}

GLEPropertyFont::GLEPropertyFont(const char* name) : GLEProperty(name, "font", GLEPropertyTypeFont, GLEDOPropertyFont) {
}

GLEPropertyFont::~GLEPropertyFont() {
}

bool GLEPropertyFont::isEqualToState(GLEPropertyStore* store) {
	int font;
	g_get_font(&font);
	GLEFont* gle_font = store->getFontProperty(this);
	return font == gle_font->getIndex();
}

void GLEPropertyFont::updateState(GLEPropertyStore* store) {
	g_set_font(store->getFontProperty(this)->getIndex());
}

GLEPropertyJustify::GLEPropertyJustify(const char* name) : GLEProperty(name, "just", GLEPropertyTypeInt, GLEDOPropertyJustify) {
}

GLEPropertyJustify::~GLEPropertyJustify() {
}

bool GLEPropertyJustify::isEqualToState(GLEPropertyStore* store) {
	int just;
	g_get_just(&just);
	return just == store->getIntProperty(this);
}

void GLEPropertyJustify::updateState(GLEPropertyStore* store) {
	g_set_just(store->getIntProperty(this));
}

void GLEPropertyJustify::getPropertyAsString(string* result, GLEPropertyValue& value) {
	int ival = value.IntValue;
	switch (ival) {
		case GLEJustifyCC:
			*result = "cc"; break;
		case GLEJustifyTL:
			*result = "tl"; break;
		case GLEJustifyTC:
			*result = "tc"; break;
		case GLEJustifyTR:
			*result = "tr"; break;
		case GLEJustifyBL:
			*result = "bl"; break;
		case GLEJustifyBC:
			*result = "bc"; break;
		case GLEJustifyBR:
			*result = "br"; break;
		case GLEJustifyLC:
			*result = "lc"; break;
		case GLEJustifyRC:
			*result = "rc"; break;
		case GLEJustifyLeft:
			*result = "left"; break;
		case GLEJustifyCenter:
			*result = "center"; break;
		case GLEJustifyRight:
			*result = "right"; break;
		default:
			*result = "?"; break;
	}
}

GLEPropertyArrowSize::GLEPropertyArrowSize(const char* name) : GLEProperty(name, "arrowsize", GLEPropertyTypeReal, GLEDOPropertyArrowSize) {
}

GLEPropertyArrowSize::~GLEPropertyArrowSize() {
}

bool GLEPropertyArrowSize::isEqualToState(GLEPropertyStore* store) {
	GLEArrowProps arrow;
	g_arrowsize(&arrow);
	return fabs(store->getRealProperty(this)-arrow.size) < 1e-18;
}

void GLEPropertyArrowSize::updateState(GLEPropertyStore* store) {
	g_set_arrow_size(store->getRealProperty(this));
}

GLEPropertyArrowAngle::GLEPropertyArrowAngle(const char* name) : GLEProperty(name, "arrowangle", GLEPropertyTypeReal, GLEDOPropertyArrowAngle) {
}

GLEPropertyArrowAngle::~GLEPropertyArrowAngle() {
}

bool GLEPropertyArrowAngle::isEqualToState(GLEPropertyStore* store) {
	GLEArrowProps arrow;
	g_arrowsize(&arrow);
	return fabs(store->getRealProperty(this)-arrow.angle) < 1e-18;
}

void GLEPropertyArrowAngle::updateState(GLEPropertyStore* store) {
	g_set_arrow_angle(store->getRealProperty(this));
}

void GLEInitColorProperty(GLEPropertyStore* prop) {
	rgb01 rgb;
	colortyp color;
	g_get_colortyp(&color);
	g_colortyp_to_rgb01(&color, &rgb);
	GLEColor* gcolor = new GLEColor(rgb.red, rgb.green, rgb.blue);
	prop->setColorProperty(GLEDOPropertyColor, gcolor);
}

void GLEInitSimpleLineProperties(GLEPropertyStore* prop) {
	// line width property
	double lwidth;
	g_get_line_width(&lwidth);
	prop->setRealProperty(GLEDOPropertyLineWidth, lwidth);
	GLEInitColorProperty(prop);
}

void GLEScaleSimpleLineProperties(double scale, bool dir, GLEPropertyStore* prop) {
	if (prop != NULL && scale > 0) {
		double wd = prop->getRealProperty(GLEDOPropertyLineWidth);
		if (dir) {
			wd *= scale;
		} else {
			wd /= scale;
		}
		prop->setRealProperty(GLEDOPropertyLineWidth, wd);
	}
}

void GLEScaleArrowProperties(double scale, bool dir, GLEPropertyStore* prop) {
	if (prop != NULL && scale > 0) {
		double size = prop->getRealProperty(GLEDOPropertyArrowSize);
		if (dir) {
			size *= scale;
		} else {
			size /= scale;
		}
		prop->setRealProperty(GLEDOPropertyArrowSize, size);
	}
}

void GLEInitLineProperties(GLEPropertyStore* prop) {
	GLEInitSimpleLineProperties(prop);
	// cap
	int cap;
	g_get_line_cap(&cap);
	prop->setIntProperty(GLEDOPropertyLineCap, cap);
}

void GLESetDefaultArrowProperties(double lwd, GLEPropertyStore* prop) {
	if (!g_is_dummy_device()) {
		g_set_dummy_device();
	}
	g_set_line_width(lwd);
	g_set_hei(0.3633);
	GLEInitArrowProperties(prop);
}

void GLEInitArrowProperties(GLEPropertyStore* prop) {
	GLEArrowProps arrow;
	g_arrowsize(&arrow);
	prop->setRealProperty(GLEDOPropertyArrowSize, arrow.size);
	prop->setRealProperty(GLEDOPropertyArrowAngle, arrow.angle);
	prop->setIntProperty(GLEDOPropertyArrowStyle, arrow.style);
	prop->setIntProperty(GLEDOPropertyArrowTip, arrow.tip);
}

